<?php

declare(strict_types=1);

namespace app\controllers;

use yii;
use yii\filters\AccessControl;
use yii\web\Controller;
use yii\web\Response;
use yii\web\Cookie;
use yii\web\UploadedFile;

use app\models\Supply;
use app\models\SupplyGroup;
use app\models\Search;
use app\models\Notification;
use app\models\Settings;
use app\models\SettingsEdgeSettings;

class ProfileController extends Controller
{
    public function behaviors()
    {
        return [
            'access' => [
                'class' => AccessControl::class,
                'rules' => [
                    [
                        'allow' => true,
                        'roles' => ['@'],
                        'actions' => ['index', 'supplies', 'import', 'monitoring', 'readGroups']
                    ],
                    [
                        'allow' => false,
                        'roles' => ['?'],
                        'denyCallback' => [$this, 'accessDenied']
                    ],
                    [
                        'allow' => true,
                        'actions' => ['panel', 'panel-notification-write'],
                        'matchCallback' => function ($rule, $action): bool {
                            if (
                                yii::$app->user->identity->type === 'administrator'
                                || yii::$app->user->identity->type === 'moderator'
                            ) {
                                return true;
                            }

                            return false;
                        }
                    ]
                ]
            ]
        ];
    }

    public function accessDenied()
    {
        // Инициализация
        $cookies = yii::$app->response->cookies;

        // Запись cookie с редиректом, который выполнится после авторизации
        $cookies->add(new Cookie([
            'name' => 'redirect',
            'value' => yii::$app->request->pathInfo
        ]));

        if (Yii::$app->request->isPost) {
            // POST-запрос

            // Настройка
            Yii::$app->response->format = Response::FORMAT_JSON;

            // Генерация ответа
            Yii::$app->response->content = json_encode([
                'main' => $this->renderPartial('/account/index'),
                'redirect' => '/profile',
                '_csrf' => Yii::$app->request->getCsrfToken()
            ]);
        } else if (Yii::$app->request->isGet) {
            // GET-запрос

            $this->redirect('/authentication');
        }
    }

    /**
     * Страница с настройками аккаунта
     */
    public function actionIndex(): string|array
    {
        // Инициализация
        $model = yii::$app->user->identity;
        $panel = yii::$app->request->post('panel') ?? yii::$app->request->get('panel');
        $sidebar = $this->renderPartial('sidebar');

        // Обработка настроек аккаунта
        if ($vars = yii::$app->request->post('Account') ?? yii::$app->request->get('Account')) {
            // Обнаружены входные параметры

            if (isset($vars['opts'])) {
                // Переданы параметры

                // Инициализация
                is_array($model->opts) || $model->opts = [];

                // Запись
                $model->opts = array_merge($model->opts, $vars['opts']);

                $model->update();
            } else {
                /**
                 * @todo Написать обработчик ошибок
                 */
            }
        }

        // Инициализация
        $delivery_from_terminal_list = $model->genListTerminalsFrom();
        $delivery_to_terminal_list = $model->genListTerminalsTo();
        $import_oem_list = $model->genListOem(Supply::searchByAccount(select: 'supply.onec["ЗначенияСвойств"]'));
        $array_unshift_in_start = function (array &$array, string|int $key, mixed $value)
        {
            $array = array_reverse($array, true);
            $array[$key] = $value;
            return $array = array_reverse($array, true);
        };
        $array_write_default_value = function (array &$array, string $key = 'Выберите', string $value = 'Выберите') use ($array_unshift_in_start) {
            if (isset($array[$key])) {
                // Смещение или ассоциация найдена

                // Деинициализация
                unset($array[$key]);

                // Инициализация
                $array_unshift_in_start($array, $key, $value);
            }
        };

        // Сортировка по алфавиту
        asort($delivery_from_terminal_list);
        asort($delivery_to_terminal_list);
        asort($import_oem_list);

        // Перемещение в начало массива значения "Выберите"
        $array_write_default_value($delivery_from_terminal_list);
        $array_write_default_value($delivery_to_terminal_list);
        $array_write_default_value($import_oem_list);

        if (yii::$app->request->isPost) {
            // POST-запрос

            yii::$app->response->format = Response::FORMAT_JSON;

            return [
                'main' => $this->renderPartial('index', compact(
                    'model',
                    'sidebar',
                    'delivery_from_terminal_list',
                    'delivery_to_terminal_list',
                    'import_oem_list',
                    'panel'
                )),
                'redirect' => '/profile',
                '_csrf' => yii::$app->request->getCsrfToken()
            ];
        }

        return $this->render('index', compact(
            'model',
            'sidebar',
            'delivery_from_terminal_list',
            'delivery_to_terminal_list',
            'import_oem_list',
            'panel'
        ));
    }

    /**
     * Страница панели управления для доверенных пользователей
     */
    public function actionPanel(): string|array
    {
        // Инициализация
        $model_notifications = null;
        $panel = yii::$app->request->post('panel') ?? yii::$app->request->get('panel');
        $sidebar = $this->renderPartial('sidebar');

        if (!$model_settings = Settings::readLast()) {
            $model_settings = new Settings();

            if(!$model_settings->save()) {
                $this->redirect('/');
            }
        }

        if ($vars = yii::$app->request->post('Notification') ?? yii::$app->request->get('Notification')) {
            // Обнаружены входные параметры из раздела "Уведомления"

            // Реинициализация с новыми параметрами
            $model_notifications = new Notification($vars);

            // Запись уведомления и отправка (запись ребра до аккаунта)
            $model_notifications->write();
        } else if ($vars = yii::$app->request->post('Settings') ?? yii::$app->request->get('Settings')) {
            // Обнаружены входные параметры из раздела "Настройки"

            if ($to = new Settings($vars)) {
                // Настройки инициализированы

                // Отправка
                if ($to->save()) {
                    // Сохранено в базе данных

                    // Буфер
                    $from = $model_settings;

                    // Реинициализация (для представления)
                    $model_settings = $to;

                    if ($from) {
                        // Найдена старая версия настроек

                        // Запись ребра: НАСТРОЙКИ (старые) -> НАСТРОЙКИ (новые)
                        SettingsEdgeSettings::write($from->readId(), $to->readId(), 'update');
                    }
                } else {
                    // Не сохранено в базе данных

                    // Запись ошибок
                    $model_settings->addErrors($to->getErrors());
                }
            }
        }

        // Деинициализация
        unset($vars);

        if (yii::$app->request->isPost) {
            // AJAX-POST-запрос

            yii::$app->response->format = Response::FORMAT_JSON;

            return [
                'main' => $this->renderPartial('panel', compact(
                    'model_notifications',
                    'model_settings',
                    'sidebar',
                    'panel'
                )),
                'redirect' => '/profile/panel',
                '_csrf' => yii::$app->request->getCsrfToken()
            ];
        }

        return $this->render('panel', compact(
            'model_notifications',
            'model_settings',
            'sidebar',
            'panel'
        ));
    }

    /**
     * Страница мониторинга
     */
    public function actionMonitoring(): string|array
    {
        // Инициализация
        $panel = yii::$app->request->post('panel') ?? yii::$app->request->get('panel');
        $sidebar = $this->renderPartial('sidebar');

        // Инициализация номера страницы
        $page_search_history = (yii::$app->request->post('search') ?? yii::$app->request->get('search')) - 1;

        if ($page_search_history <= 0) {
            // Вышли за границу поиска перед первой страницей

            // Инициализация
            $page_search_history = 0;
        }

        // Инициализация количества строк на одной странице
        $rows_amount = 10;

        $search_history = Search::SearchByEdge(
            from: 'account',
            to: 'search',
            subquery_where: [
                [
                    'account._id' => yii::$app->user->id
                ]
            ],
            foreach: ['edge' => 'account_edge_search'],
            where: 'edge._to == search._id',
            limit: $rows_amount,
            offset: $offset = ((int) $page_search_history ?? 0) * $rows_amount
        );

        // Проверка результатов
        monitoring_result_check:

        if (count($search_history) === 0 && $offset !== 0) {
            // Вышли за границу поиска после последней страницы

            // Реинициализация (вычитание для идентичного конечного результата)
            --$page_search_history;

            $search_history = Search::SearchByEdge(
                from: 'account',
                to: 'search',
                subquery_where: [
                    [
                        'account._id' => yii::$app->user->id
                    ]
                ],
                foreach: ['edge' => 'account_edge_search'],
                where: 'edge._to == search._id',
                limit: $rows_amount,
                offset: $offset = ((int) $page_search_history ?? 0) * $rows_amount
            );

            // Повторная проверка
            goto monitoring_result_check;
        }

        if (yii::$app->request->isPost) {
            // AJAX-POST-запрос

            yii::$app->response->format = Response::FORMAT_JSON;

            return [
                'main' => $this->renderPartial('monitoring', compact(
                    'sidebar',
                    'search_history',
                    'page_search_history',
                    'panel'
                )),
                'search' => $page_search_history + 1,
                'redirect' => '/profile/monitoring',
                '_csrf' => yii::$app->request->getCsrfToken()
            ];
        }

        return $this->render('monitoring', compact(
            'sidebar',
            'search_history',
            'page_search_history',
            'panel'
        ));
    }

    /**
     * Страница поставок
     */
    public function actionSupplies(): string|array
    {
        // Инициализация
        $model = new Supply(yii::$app->request->post('Supply') ?? yii::$app->request->get('Supply'));
        $panel = yii::$app->request->post('panel') ?? yii::$app->request->get('panel');
        $sidebar = $this->renderPartial('sidebar');
        $groups = self::readGroups();


        if (yii::$app->request->isPost) {
            // POST-запрос

            yii::$app->response->format = Response::FORMAT_JSON;

            return [
                'main' => $this->renderPartial('supplies', compact(
                    'model',
                    'groups',
                    'sidebar',
                    'panel'
                )),
                'redirect' => '/profile/supplies',
                '_csrf' => yii::$app->request->getCsrfToken()
            ];
        }

        return $this->render('supplies', compact(
            'model',
            'groups',
            'sidebar',
            'panel'
        ));
    }

    /**
     * Импорт поставок
     *
     * На данный момент только из Excel-таблицы
     */
    public function actionImport()
    {
        // Инициализация
        $model = new Supply(yii::$app->request->post('Supply') ?? yii::$app->request->get('Supply'));
        $model->scenario = $model::SCENARIO_IMPORT_EXCEL;
        $panel = yii::$app->request->post('panel') ?? yii::$app->request->get('panel');
        $sidebar = $this->renderPartial('sidebar');
        $groups = self::readGroups();

        if (yii::$app->request->isPost) {
            // AJAX-POST-запрос

            yii::$app->response->format = Response::FORMAT_JSON;

            $model->file_excel = UploadedFile::getInstances($model, 'file_excel');

            if ($model->importExcel()) {
                return [
                    'main' => $this->renderPartial('supplies', compact(
                        'model',
                        'groups',
                        'sidebar',
                        'panel'
                    )),
                    '_csrf' => yii::$app->request->getCsrfToken()
                ];
            }

            yii::$app->response->statusCode = 409;
        }

        return $this->render('supplies', compact(
            'model',
            'groups',
            'sidebar',
            'panel'
        ));
    }

    public static function readGroups()
    {
        // Инициализация
        $groups = [];

        foreach (SupplyGroup::readAll() as $group) {
            // Перебор всех групп

            // Генерация [КЛЮЧ => ИМЯ]
            $groups[$group->_key] = $group->name;
        }

        return $groups;
    }
}
